% Testing Davenport's (eigen) and Markley's (SVD) solutions to Wahba's
% problem: To get a Least Squares Estimate of Spacecraft Attitude"
% (Wahba's problem was published in 1965, SIAM Review, Vol 7, No 3)

% Wahba's problem was first solved by Davenport in 1971. And later by
% Schuster and Markley around 1988.

% The two suggested algorithms herein are tested with noiseless observations 
% and with a random attitude induced on the vehicle body.

% The main source document used was "Humble Problems" by F. Landis Markley - 2006
% Freely downloadable at the time of writing from:
% http://ntrs.nasa.gov/archive/nasa/casi.ntrs.nasa.gov/20060012294_2006013132.pdf

% V1.2
% Refinements 01 and 02 was done by consulting the original Markley article:
% "Attitude Determination using Observations and the SVD" - Journal of 
% Astronatical Science, Vol 38, No 3, July-Sept 1998, p 245-258

% Wabha's problem in short is determining ones (the body's) attitude using
% a number or co-registered vectors in a reference frame and observations 
% in the body coordinates:

% Basically the problem is minimizing the following cost function to get R, 
% the rotation matrix (or attitude quaternion/dcm) : 
%        L = 0.5 SUM a_i (b_i - R r_i)^2                    Eq(1)
% where
%        a_i - is the weights (a below)
%        b_i - observations in body coordinates (rb below)
%        r_i - known database coregistered datapoints in a reference
%        coordinates (rr below)
%
% the above is equivalent to solving in quaternion from:
%        L = lambda_0 - trace(RB) = lambda_0 - q' K q       Eq(2)
% where
%        q - the attitude quaternion
%        K - is calculated as below
%
% Please follow the code. You will see the Equations were simply 
% implemented as in the article. There is one change however, the author 
% preferred using Zipfels order for the quaternion representation, thus:
% q = [ q0 q1 q2 q3 ] = [ cos (the/2) e(1)sin(the/2) e(1)sin(the/2) e(1)sin(the/2) ]

% Note that Body orienations can similarly be found with Gupta's 1998 method or
% by simply co-registering a known star-database distance matrix based on observed 
% angular-distances. Note that Gupta's work seems to be based on Hyslop 1987 work. 

% See also Julian Balfour's work in on three-point tracking (basically a 
% distance matrix traversing method). 

% Copyright (C) 2013 - Pieter V. Reyneke, South Africa

% Open Source licence, use whereever you like but at own risk, keep this
% complete copyright notice in the code and please send me updates and 
% report errors to pieter.reyneke@gmail.com. Will give recognition, if
% requested, after any valid comment leading to an update was recieved. 
%
% Author: PV Reyneke

clear variables;
close all;

figure_num = 11;
hh = figure(figure_num);
whitebg(hh,'w');
colormap('default'); 
axis off;
set(hh, 'name', ['Solution to Wahba']);
rotate3d;
    
window_docked = 1;
if (window_docked)
    set(hh, 'Windowstyle', 'docked');
else
    cameratoolbar('Show');
    cameratoolbar('SetMode','orbit','SetCoordSys','none');
end
%% The number of random observations to create
no = 300;

% Lets get an average performance over a number or iterations
niter = 9;

% Choose an orientation for the vehicle body randomly (including roll)
n = 1;
by = 1.*(rand(n,1).*2*pi-pi)+0*deg2rad(37);
bp = 1.*(asin(rand(n,1).*2-1))+0*deg2rad(0);
br = 1.*(rand(n,1).*2*pi-pi)+0*pi+0*deg2rad(0);

% Make a quaternion of it and turn the body by the quaternion
% bq = d2q(e2d_ypr([by,bp,br]));
bq = qconj(e2q_ypr([by,bp,br]));
eq = q2e_ypr(qconj(bq));

sol = zeros(niter,6);
display(['Iter:Num = ( TimeDavenport     TimeMarkley       ErrDavenport      ErrMarkley        ErrDav-ErrMarkley Dist(Dav,Markley))'])
for j=1:niter
    
    % Create a number of random (directions over a sphere) observations vectors [x y z]
    oy = 1.*(rand(no,1).*2*pi-pi);
    op = 1.*(asin(rand(no,1).*2-1));
    or = 1.*(rand(no,1).*2e12)+1;
    [x y z] = sph2cart(oy,op,or);
    [xu yu zu] = sph2cart(oy,op,0.*or+1);
    ru = [xu yu zu]';
    rr = [x y z]';

    % Turn the observations to observations in the body frame
    rb = rotq(bq,rr);
    % rb = rr;  % to test unity

    % Allocate non-negative random weights 
    % (note weights does not matter here because there is no noise on the obs)
    a = 1.*rand([no,1])+0;
    
    % Refinement 01 - Normalise the weights
    a = a./sum(a);

    % Calc lambda0, think it was t check (Eq13b Humble Problems)
    lambda0 = sum(a);
    B = ([a a a]'.*rb)*rr';          % Eq13b Humble Problems
    BT = transpose(B);          % Eq13b Humble Problems
    K = [                            % Eq13b Humble Problems  
            trace(B) sum(cross([a a a].*rb',rr',2))          
            sum(cross([a a a].*rb',rr',2))' B+B'-eye(3).*trace(B) 
        ];

    tic;
    [V,D] = eig(K,'nobalance');      % Eq14 Humble Problems - Davenport's Solution
    [d id] = max(diag(D));
    qopt = V(:,id)';                  % Eq14 Humble Problems - Davenport's Solution
    t1 = toc;
    
    tic;
    [U2 S2 V2] = svd(BT);            % Eq14 Humble Problems - Markleys Solution
                                     % What Marley does not say is that the
                                     % SVD of the transpose should be taken
                                     
    % Refinement 02
    do_refine_02 = 1;
    if (do_refine_02)
        U_plus = U2*diag([1 1 det(U2)]); 
        V_plus = V2*diag([1 1 det(V2)]); 
        d = det(U2)*det(V2);
        d_plus = det(U_plus)*det(V_plus);
        S123 = diag(S2);
        S_plus = diag([S123(1) S123(2) d*S123(3)]);
        % The idea of this refinement was to connect the orthogonal solution to 
        % the Rodigues params as follows (i.e. breaking up a proper orthogonal 
        % matrix into e and Phi) see Toussaint's writings also:
        % W = (U_plus'*(R*V_plus)) = cos(Phi)*eye(3) + (1-cos(Phi))e*e' - sin(Phi)*skew(e)
        % if we had R we could do the following:
        % W = (U_plus'*(R*V_plus));
        % [e Phi] = rotn(W);    
        
        % This can then be subtituted into Eq(1) above to get:
        % L(R) = 1 - trace(S_plus*W)
        %      = 1 - tr(S_plus) + (1-cos(Phi))[S123(2)+d*S123(3)+
        %        (S123(1)+S123(2))e(2)^2+(S123(1)-d*S123(3)e(3)^2)]
        % Then it follows that:
        Ropt = U_plus*V_plus';        % Eq18 Markley Wahba SVD 1998 article
        qopt2 = d2q(Ropt);            % gives the equivalent Markleys Solution 

        % Check - the following should give close to zero
        [BT-(U_plus*(S_plus*V_plus'))];% Eq13 Markley Wahba SVD 1998 - 
    else
        % or The 2006 SVD Solution
        Ropt = U2*(diag([1 1 det(U2)*det(V2)])*V2');
        qopt2 = d2q(Ropt);            % Eq14 Humble Problems - Markleys Solution 
        
        % Check - the following should give close to zero
        [BT-(U2*(S2*V2'))];           % Eq15a Humble Problems 
    end
    t2 = toc;
    
    ecalc = rad2deg(q2e_ypr(qconj(qopt2)));
    eorig = rad2deg([by, bp, br]); 

    % Rtrue = e2d_ypr([by, bp, br]); 
    Rtrue = q2d(d2q(q2d(bq))); 
    
    set(0, 'CurrentFigure', hh)
    XX = Ropt'*[1;0;0];
    YY = Ropt'*[0;1;0];
    ZZ = Ropt'*[0;0;1];
    XT = Rtrue'*[1;0;0];
    YT = Rtrue'*[0;1;0];
    ZT = Rtrue'*[0;0;1];
    % XT = rotq(bq,[1;0;0]);
    % YT = rotq(bq,[0;1;0]);
    % ZT = rotq(bq,[0;0;1]);
    ccplot3(xu, yu, zu, or); hold on;
    arrow3D([0 0 0],[1 0 0],0.9,[0 0 1]);
    arrow3D([0 0 0],[0 1 0],0.9,[0 0 1]);
    arrow3D([0 0 0],[0 0 1],0.9,[0 0 1]);
    arrow3D([0 0 0],XT,0.85,[0 1 0]);
    arrow3D([0 0 0],YT,0.85,[0 1 0]);
    arrow3D([0 0 0],ZT,0.85,[0 1 0]);
    arrow3D([0 0 0],XX,0.9,[1 0 0]);
    arrow3D([0 0 0],YY,0.9,[1 0 0]);
    arrow3D([0 0 0],ZZ,0.9,[1 0 0]);
    light 
    lighting phong; %lighting gouraud;
    xlabel('x');ylabel('y');zlabel('z');
    axis off; axis tight;
    axis equal; %grid on;  axis vis3d;
    hold off;
    if (~window_docked)
       figsize(hh,[0.9 0.9]);
    end
    
    e1 = abs(quat_dist(bq,qopt));
    e2 = abs(quat_dist(bq,qopt2));
    title({['\fontsize{14}Rotation R_{yrp} [true](calculated)'],...
               ['\fontsize{12}[' num2str(eorig,' %+07.4f') '](' num2str(ecalc,' %+07.4f') ')'],...
               ['\fontsize{8} '],...
               ['\fontsize{10}=> e_{davenport}:' num2str(e1,' %+017.14f') ', e_{markley}:' num2str(e2,' %+017.14f')]});
    drawnow;
    
    % Compare Davenport to qopt, Markley to qopt and Davenport to Markley
    sol(j,:) = [t1 t2 quat_dist(bq,qopt) quat_dist(bq,qopt2) quat_dist(bq,qopt)-quat_dist(bq,qopt2) quat_dist(qopt,qopt2) ];
    display(['Iter:' num2str(j,'%03d') ' = (' num2str(real(sol(j,:)),' %+017.14f') ')'])
end
display(['Iter:Num = (TimeDavenport(avg) TimeMarkley(avg)  ErrDavenport(avg) ErrMarkley(avg) ErrDav-ErrMarkley(avg) Dist(Dav,Markley)(avg))'])
display(['Mean:All = (' num2str(mean(abs(sol)),'%+017.14f ') ')'])




